.highlight_selected_word_toggle > .fa {
	margin-top: -2px;
	margin-right: -16px;
}

.notebook_app.edit_mode .CodeMirror:not(.CodeMirror-focused) .cm-matchhighlight:not(.CodeMirror-selectedtext),
.notebook_app.edit_mode .CodeMirror:not(.CodeMirror-focused) .CodeMirror-selection-highlight-scrollbar {
	background-color: #BBFFBB; /* this color gets overwritten by js anyway */
}

.notebook_app.edit_mode .CodeMirror.CodeMirror-focused .cm-matchhighlight:not(.CodeMirror-selectedtext),
.notebook_app.edit_mode .CodeMirror.CodeMirror-focused .CodeMirror-selection-highlight-scrollbar {
    background-color:#90EE90; /* this color gets overwritten by js anyway */
}

.notebook_app.edit_mode .CodeMirror .cm-matchhighlight-outline {
	outline-style: solid;
	outline-width: 2px; /* this value gets overwritten by js anyway */
	outline-color: #BBFFBB; /* this color gets overwritten by js anyway */
}

.notebook_app.edit_mode .CodeMirror.CodeMirror-focused .cm-matchhighlight-outline {
    outline-color:#90EE90; /* this color gets overwritten by js anyway */
}
